@@ -19,7 +19,6 @@ import subprocess
 import csv
 import os
 import psycopg2
-import six
 from pathlib import Path
 from datasize import DataSize
 
@@ -35,7 +34,7 @@ else:
 
 MAX_CONTENT_LENGTH = web.app.config.get('MAX_CONTENT_LENGTH') or 10485760
 QSV_BIN = web.app.config.get('QSV_BIN') or '/usr/local/bin/qsvdp'
-PREVIEW_ROWS = web.app.config.get('PREVIEW_ROWS') or 10000
+PREVIEW_ROWS = web.app.config.get('PREVIEW_ROWS') or 1000
 DEFAULT_EXCEL_SHEET = web.app.config.get('DEFAULT_EXCEL_SHEET') or 0
 CHUNK_SIZE = web.app.config.get('CHUNK_SIZE') or 16384
 DOWNLOAD_TIMEOUT = web.app.config.get('DOWNLOAD_TIMEOUT') or 30
@@ -67,14 +66,19 @@ else:
 if not SSL_VERIFY:
     requests.packages.urllib3.disable_warnings()
 
+POSTGRES_INT_MAX = 2147483647
+POSTGRES_INT_MIN = -2147483648
+POSTGRES_BIGINT_MAX = 9223372036854775807
+POSTGRES_BIGINT_MIN = -9223372036854775808
+
 _TYPE_MAPPING = {
     'String': 'text',
-    # 'int' may not be big enough,
-    # and type detection may not realize it needs to be big
-    'Integer': 'numeric',
+    # smartint looks at the min/max values of qsv stats scan and
+    # chooses integer, bigint or numeric as appropriate
+    'Integer': 'smartint',
     'Float': 'numeric',
     'DateTime': 'timestamp',
-    'Date': 'timestamp',
+    'Date': 'date',
     'NULL': 'text',
 }
 
@@ -357,6 +361,13 @@ def push_to_datastore(task_id, input, dry_run=False):
     logger.addHandler(handler)
     logger.setLevel(logging.DEBUG)
 
+    # check if QSV_BIN exists
+    qsv_path = Path(QSV_BIN)
+    if not qsv_path.is_file():
+        raise util.JobError(
+            '{} not found.'.format(QSV_BIN)
+        )
+
     validate_input(input)
 
     data = input['metadata']
@@ -447,9 +458,25 @@ def push_to_datastore(task_id, input, dry_run=False):
 
     resource['hash'] = file_hash
 
-    # Start Analysis using qsv instead of messytables, as 1) its type inferences are bullet-proof
-    # not guesses as it scans the entire file, 2) its super-fast, and 3) it has
-    # addl data-wrangling capabilities we use in datapusher+ - slice, input, count, headers, etc.
+    def cleanup_tempfiles():
+        # cleanup temporary files
+        if os.path.exists(tmp.name + ".idx"):
+            os.remove(tmp.name + ".idx")
+        tmp.close()
+        if 'qsv_slice_csv' in globals():
+            qsv_slice_csv.close()
+        if 'qsv_excel_csv' in globals():
+            qsv_excel_csv.close()
+        if 'qsv_input_csv' in globals():
+            qsv_input_csv.close()
+        if 'qsv_dedup_csv' in globals():
+            qsv_dedup_csv.close()
+
+    '''
+    Start Analysis using qsv instead of messytables, as 1) its type inferences are bullet-proof
+    not guesses as it scans the entire file, 2) its super-fast, and 3) it has
+    addl data-wrangling capabilities we use in datapusher+ - slice, input, count, headers, etc.
+    '''
     fetch_elapsed = time.perf_counter() - timer_start
     logger.info('Fetched {:.2MB} file in {:,.2f} seconds. Analyzing with qsv...'.format(
         DataSize(cl), fetch_elapsed))
@@ -460,38 +487,47 @@ def push_to_datastore(task_id, input, dry_run=False):
     format = resource.get('format').upper()
     if format in spreadsheet_extensions:
         # if so, export it as a csv file
-        logger.info('Converting {} sheet {} to CSV...'.format(
-            format, DEFAULT_EXCEL_SHEET))
-        # first, we need a temporary spreadsheet filename with the right file extension
-        # we only need the filename though, that's why we remove/delete it
-        # and create a hardlink to the file we got from CKAN
+        logger.info('Converting {} to CSV...'.format(format))
+        '''
+        first, we need a temporary spreadsheet filename with the right file extension
+        we only need the filename though, that's why we remove it
+        and create a hardlink to the file we got from CKAN
+        '''
         qsv_spreadsheet = tempfile.NamedTemporaryFile(suffix='.' + format)
         os.remove(qsv_spreadsheet.name)
         os.link(tmp.name, qsv_spreadsheet.name)
 
         # run `qsv excel` and export it to a CSV
+        # use --trim option to trim column names and the data
         qsv_excel_csv = tempfile.NamedTemporaryFile(suffix='.csv')
         try:
             qsv_excel = subprocess.run(
-                [QSV_BIN, 'excel', qsv_spreadsheet.name, '--sheet', str(DEFAULT_EXCEL_SHEET), '--output', qsv_excel_csv.name], check=True)
+                [QSV_BIN, 'excel', qsv_spreadsheet.name, '--sheet', str(DEFAULT_EXCEL_SHEET),
+                 '--trim', '--output', qsv_excel_csv.name],
+                 check=True, capture_output=True, text=True)
         except subprocess.CalledProcessError as e:
-            tmp.close()
+            cleanup_tempfiles()
             qsv_excel_csv.close()
             raise util.JobError(
                 'Cannot export spreadsheet to CSV: {}'.format(e)
             )
         qsv_spreadsheet.close()
+        excel_export_msg = qsv_excel.stderr
+        logger.info("{}...".format(excel_export_msg))
         tmp = qsv_excel_csv
     else:
-        # its a CSV/TSV? Check if its valid and
-        # transcode it to UTF-8 at the same time
+        '''
+        its a CSV/TSV? Check if its RFC4180 valid & transcode to UTF-8 at the same time
+        if its a TSV/TAB file, convert to CSV as well for standard handling downstream
+        note that we only change the working file, the resource file itself is unchanged
+        '''
         qsv_input_csv = tempfile.NamedTemporaryFile(suffix='.csv')
         logger.info('Validating/Transcoding {}...'.format(format))
         try:
             qsv_input = subprocess.run(
                 [QSV_BIN, 'input', tmp.name, '--output', qsv_input_csv.name], check=True)
         except subprocess.CalledProcessError as e:
-            tmp.close()
+            cleanup_tempfiles()
             qsv_input_csv.close()
             raise util.JobError(
                 'Invalid CSV file: {}'.format(e)
@@ -499,6 +535,7 @@ def push_to_datastore(task_id, input, dry_run=False):
         tmp = qsv_input_csv
 
     # do we need to dedup?
+    # note that deduping also ends up creating a sorted CSV
     if QSV_DEDUP:
         qsv_dedup_csv = tempfile.NamedTemporaryFile(suffix='.csv')
         logger.info('Checking for duplicate rows...')
@@ -506,7 +543,7 @@ def push_to_datastore(task_id, input, dry_run=False):
             qsv_dedup = subprocess.run(
                 [QSV_BIN, 'dedup', tmp.name, '--output', qsv_dedup_csv.name], capture_output=True, text=True)
         except subprocess.CalledProcessError as e:
-            tmp.close()
+            cleanup_tempfiles()
             qsv_dedup_csv.close()
             raise util.JobError(
                 'Check for duplicates error: {}'.format(e)
@@ -514,7 +551,7 @@ def push_to_datastore(task_id, input, dry_run=False):
         dupe_count = int(str(qsv_dedup.stderr).strip())
         if dupe_count > 0:
             tmp = qsv_dedup_csv
-            logger.info(
+            logger.warning(
                 '{:,} duplicates found and removed...'.format(dupe_count))
         else:
             logger.info('No duplicates found...')
@@ -523,9 +560,9 @@ def push_to_datastore(task_id, input, dry_run=False):
     # are all accelerated/multithreaded when an index is present
     try:
         subprocess.run(
-            [QSV_BIN, 'index', tmp.name], capture_output=True)
+            [QSV_BIN, 'index', tmp.name], check=True)
     except subprocess.CalledProcessError as e:
-        tmp.close()
+        cleanup_tempfiles
         raise util.JobError(
             'Cannot index CSV: {}'.format(e)
         )
@@ -533,24 +570,33 @@ def push_to_datastore(task_id, input, dry_run=False):
     # get record count
     try:
         qsv_count = subprocess.run(
-            [QSV_BIN, 'count', tmp.name], capture_output=True, text=True)
+            [QSV_BIN, 'count', tmp.name], capture_output=True, check=True, text=True)
     except subprocess.CalledProcessError as e:
-        tmp.close()
+        cleanup_tempfiles()
         raise util.JobError(
             'Cannot count records in CSV: {}'.format(e)
         )
     record_count = int(str(qsv_count.stdout).strip())
-    logger.info('{:,} records detected...'.format(record_count))
+    unique_qualifier = ''
+    if QSV_DEDUP:
+        unique_qualifier = 'unique'
+    logger.info('{:,} {} records detected...'.format(
+        record_count, unique_qualifier))
 
-    # if DATELIKE_FIELDNAMES is not empty, scan CSV headers for date-like field,
-    # otherwise, always --infer-dates when scanning for types
+    '''
+    if DATELIKE_FIELDNAMES is not empty, scan CSV headers for date-like field,
+    otherwise, always --infer-dates when scanning for types
+    we do this since date type inference is a very expensive op
+    as qsv needs to check if a string is a date using 15 date formats and permutations thereof
+    exponentially slowing down qsv, even if its using SIMD-accelerated, multithreaded regex parsing
+    '''
     inferdates_flag = True
     if DATELIKE_FIELDNAMES:
         try:
             qsv_headers = subprocess.run(
-                [QSV_BIN, 'headers', tmp.name], capture_output=True, text=True)
+                [QSV_BIN, 'headers', '--just-names', tmp.name], capture_output=True, check=True, text=True)
         except subprocess.CalledProcessError as e:
-            tmp.close()
+            cleanup_tempfiles()
             raise util.JobError(
                 'Cannot scan CSV headers: {}'.format(e)
             )
@@ -561,26 +607,33 @@ def push_to_datastore(task_id, input, dry_run=False):
     # run qsv stats to get data types and descriptive statistics
     headers = []
     types = []
+    headers_min = []
+    headers_max = []
     qsv_stats_csv = tempfile.NamedTemporaryFile(suffix='.csv')
     qsv_stats_cmd = [QSV_BIN, 'stats', tmp.name,
                      '--output', qsv_stats_csv.name]
     if inferdates_flag:
         qsv_stats_cmd.append('--infer-dates')
-        logger.info('Date-like fields detected. Date inferencing enabled...')
+        # only show the log message when smart date inferencing is on
+        # when smart date inferencing is off, we always infer dates
+        if DATELIKE_FIELDNAMES:
+            logger.info(
+                'Date-like fields detected. Date inferencing enabled...')
     try:
         qsv_stats = subprocess.run(qsv_stats_cmd, check=True)
     except subprocess.CalledProcessError as e:
-        tmp.close()
+        cleanup_tempfiles()
         qsv_stats_csv.close()
         raise util.JobError(
             'Cannot infer data types and compile statistics: {}'.format(e)
         )
     with open(qsv_stats_csv.name, mode='r') as inp:
-        reader = csv.reader(inp)
-        next(reader)  # skip first element, which is a header
-        for rows in reader:
-            headers.append(rows[0])
-            types.append(rows[1])
+        reader = csv.DictReader(inp)
+        for row in reader:
+            headers.append(row['field'])
+            types.append(row['type'])
+            headers_min.append(row['min'])
+            headers_max.append(row['max'])
 
     existing = datastore_resource_exists(resource_id, api_key, ckan_url)
     existing_info = None
@@ -592,7 +645,7 @@ def push_to_datastore(task_id, input, dry_run=False):
     if existing_info:
         types = [{
             'text': 'String',
-            'numeric': 'Decimal',
+            'numeric': 'Float',
             'timestamp': 'DateTime',
         }.get(existing_info.get(h, {}).get('type_override'), t)
             for t, h in zip(types, headers)]
@@ -607,8 +660,23 @@ def push_to_datastore(task_id, input, dry_run=False):
             res_id=resource_id))
         delete_datastore_resource(resource_id, api_key, ckan_url)
 
-    headers_dicts = [dict(id=field[0], type=TYPE_MAPPING[str(field[1])])
-                     for field in zip(headers, types)]
+    # 1st pass of building headers_dict
+    temp_headers_dicts = [dict(id=field[0], type=TYPE_MAPPING[str(field[1])])
+                          for field in zip(headers, types)]
+
+    # 2nd pass header_dicts, checking for smartint types
+    headers_dicts = []
+    for idx, header in enumerate(temp_headers_dicts):
+        if header['type'] == 'smartint':
+            if int(headers_max[idx]) <= POSTGRES_INT_MAX and int(headers_min[idx]) >= POSTGRES_INT_MIN:
+                header_type = 'integer'
+            elif int(headers_max[idx]) <= POSTGRES_BIGINT_MAX and int(headers_min[idx]) >= POSTGRES_BIGINT_MIN:
+                header_type = 'bigint'
+            else:
+                header_type = 'numeric'
+        else:
+            header_type = header['type']
+        headers_dicts.append(dict(id=header['id'], type=header_type))
 
     # Maintain data dictionaries from matching column names
     if existing_info:
@@ -633,7 +701,7 @@ def push_to_datastore(task_id, input, dry_run=False):
             qsv_slice = subprocess.run(
                 [QSV_BIN, 'slice', '--len', str(PREVIEW_ROWS), tmp.name, '--output', qsv_slice_csv.name], check=True)
         except subprocess.CalledProcessError as e:
-            tmp.close()
+            cleanup_tempfiles()
             qsv_slice_csv.close()
             raise util.JobError(
                 'Cannot create a preview slice: {}'.format(e)
@@ -655,21 +723,26 @@ def push_to_datastore(task_id, input, dry_run=False):
     send_resource_to_datastore(resource, headers_dicts, api_key, ckan_url,
                                records=None, aliases=None, calculate_record_count=False)
 
-    record_count = 0
+    copied_count = 0
     try:
         raw_connection = psycopg2.connect(WRITE_ENGINE_URL)
     except psycopg2.Error as e:
-        tmp.close()
+        cleanup_tempfiles()
         raise util.JobError(
             'Could not connect to the Datastore: {}'.format(e)
         )
     else:
         cur = raw_connection.cursor()
-        # truncate table to use copy freeze option and further increase
-        # performance as there is no need for WAL logs to be maintained
-        # https://www.postgresql.org/docs/9.1/populate.html#POPULATE-COPY-FROM
-        cur.execute('TRUNCATE TABLE \"{resource_id}\";'.format(
-            resource_id=resource_id))
+        '''
+        truncate table to use copy freeze option and further increase
+        performance as there is no need for WAL logs to be maintained
+        https://www.postgresql.org/docs/9.1/populate.html#POPULATE-COPY-FROM
+        '''
+        try:
+            cur.execute('TRUNCATE TABLE \"{resource_id}\";'.format(
+                resource_id=resource_id))
+        except psycopg2.Error as e:
+            logger.warning("Could not TRUNCATE: {}".format(e))
 
         copy_sql = ("COPY \"{resource_id}\" ({column_names}) FROM STDIN "
                     "WITH (FORMAT CSV, FREEZE 1, "
@@ -682,24 +755,30 @@ def push_to_datastore(task_id, input, dry_run=False):
             try:
                 cur.copy_expert(copy_sql, f)
             except psycopg2.Error as e:
-                logger.warning("Postgres COPY failed: {}".format(e))
+                cleanup_tempfiles()
+                raise util.JobError(
+                    'Postgres COPY failed: {}'.format(e)
+                )
             else:
-                record_count = cur.rowcount
+                copied_count = cur.rowcount
 
         raw_connection.commit()
         # this is needed to issue a VACUUM ANALYZE
         raw_connection.set_isolation_level(
             psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
         cur = raw_connection.cursor()
-        logger.info('Vacuum Analyzing table...')
         cur.execute('VACUUM ANALYZE \"{resource_id}\";'.format(
             resource_id=resource_id))
 
     copy_elapsed = time.perf_counter() - copy_start
     logger.info('...copying done. Copied {n} rows to "{res_id}" in {copy_elapsed} seconds.'.format(
-        n='{:,}'.format(record_count), res_id=resource_id, copy_elapsed='{:,.2f}'.format(copy_elapsed)))
+        n='{:,}'.format(copied_count), res_id=resource_id, copy_elapsed='{:,.2f}'.format(copy_elapsed)))
 
     resource['datastore_active'] = True
+    resource['total_record_count'] = record_count
+    if PREVIEW_ROWS < record_count:
+        resource['preview'] = True
+        resource['preview_rows'] = copied_count
     update_resource(resource, api_key, ckan_url)
 
     if AUTO_ALIAS:
@@ -713,7 +792,7 @@ def push_to_datastore(task_id, input, dry_run=False):
             owner_org_name = owner_org.get('name')
         if resource_name and package_name and owner_org_name:
             alias = f"{resource_name}-{package_name}-{owner_org_name}"[:59]
-            # check if the alias exist, if it does
+            # check if the alias already exist, if it does
             # add a sequence suffix so the new alias can be created
             cur.execute('SELECT COUNT(*) FROM _table_metadata where name like \'{}%\';'.format(
                 alias))
@@ -731,6 +810,8 @@ def push_to_datastore(task_id, input, dry_run=False):
                         break
                     alias_sequence += 1
         else:
+            logger.info(
+                'Cannot create alias: {}-{}-{}'.format(resource_name, package_name, owner_org))
             alias = None
     raw_connection.close()
 
@@ -740,18 +821,7 @@ def push_to_datastore(task_id, input, dry_run=False):
     if alias:
         logger.info('Created alias: {}'.format(alias))
 
-    # cleanup temporary files
-    if os.path.exists(tmp.name + ".idx"):
-        os.remove(tmp.name + ".idx")
-    tmp.close()
-    if 'qsv_slice_csv' in globals():
-        qsv_slice_csv.close()
-    if 'qsv_excel_csv' in globals():
-        qsv_excel_csv.close()
-    if 'qsv_input_csv' in globals():
-        qsv_input_csv.close()
-    if 'qsv_dedup_csv' in globals():
-        qsv_dedup_csv.close()
+    cleanup_tempfiles()
 
     total_elapsed = time.perf_counter() - timer_start
     logger.info(
