@@ -1,5 +1,4 @@
 import os
-import six
 import ckanserviceprovider.web as web
 
 from datapusher import jobs
@@ -22,15 +21,11 @@ def main():
     import argparse
 
     argparser = argparse.ArgumentParser(
-        description='Service that allows automatic migration of data to the CKAN DataStore',
-        epilog='''"He reached out and pressed an invitingly large red button on a nearby panel.
-                The panel lit up with the words Please do not press this button again."''')
-    if six.PY3:
-        argparser.add_argument('config', metavar='CONFIG', type=argparse.FileType('r'),
-                            help='configuration file')
-    if six.PY2:
-        argparser.add_argument('config', metavar='CONFIG', type=file,
-                            help='configuration file')
+        description='Service that allows automatic migration of data to the CKAN DataStore')
+
+    argparser.add_argument('config', metavar='CONFIG', type=argparse.FileType('r'),
+                           help='configuration file')
+
     args = argparser.parse_args()
 
     os.environ['JOB_CONFIG'] = os.path.abspath(args.config.name)
@@ -41,10 +36,10 @@ def main():
 def initdb():
     import argparse
 
-    argparser = argparse.ArgumentParser(description='Initializes de database')
+    argparser = argparse.ArgumentParser(description='Initializes the database')
 
     argparser.add_argument('config', metavar='CONFIG', type=argparse.FileType('r'),
-                            help='configuration file')
+                           help='configuration file')
 
     args = argparser.parse_args()
 
